package choco;

import org.chocosolver.solver.Model;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.util.tools.ArrayUtils;
import org.chocosolver.solver.Solver;
import static org.chocosolver.solver.search.strategy.Search.*;

//import static org.chocosolver.solver.search.strategy.Search.minDomLBSearch;
/*
 * A group of nf friends is organising a number of nw weekly lunches on a restaurant, 
 * which has a round table that fit all of them (i.e. it seats nf people). 
 * Because the goal is to socialise, the organiser is trying to find an allocation 
 * of seats for the participants, so that each of them has different neighbours every week.
 */
public class round_table {
	
	public static void main(String[] args) {
		int nw = 4;
		int nf = 9;
		boolean first_fail = true;
		
		Model model = new Model("Round Table (" + nf + "," + nw + ")");
		Solver s = model.getSolver();
		
		// pos[w][f] represents the position of friend f in week w
		IntVar[][] pos = model.intVarMatrix("T", nw, nf, 0, nf-1);
		
		// dst[map[w][f][g]] represents the distance between friends f and g in week w
		IntVar[] dst = model.intVarArray("D", nw*nf*(nf-1)/2, 0, nf-1);
		int [][][] map = new int [nw][nf][nf];
		
		// map triples <w,f,g> into an index of vector dst;
		int k = 0;
		for (int w = 0; w < nw; w++){
			for(int f = 0; f < nf-1; f++){
				for (int g = f+1; g < nf; g++){
					map[w][f][g] = k;
					k = k + 1;
				}
			}
		}
		
		// In the first week, the positions are arbitrary
		for(int f=0; f < nf; f++){
			model.arithm(pos[0][f] ,"=", f).post();
		}
		
		// in every week, all friends are at the table
		for(int w = 1; w < nw; w++) {
			model.allDifferent(pos[w], "AC").post();
		}
		

		// constrain the distance of any two friends in any week
		k = 0;
		for(int w = 0; w < nw; w++){
			for(int f1 = 0; f1 < nf-1; f1++){
				for(int f2 = f1+1; f2 < nf; f2++){
					//System.out.println(k + " : (" + w +"," + f1 +"," + f2 + ")");
					dst[k].eq(pos[w][f1].sub(pos[w][f2]).abs()
							 .min((pos[w][f1].sub(pos[w][f2]).add(nf)).abs())
							 .min((pos[w][f1].sub(pos[w][f2]).sub(nf)).abs())
						   ).post();
					k = k + 1;
				}
			}
		}
		//System.out.println();
		
		// For every two weeks, at least one of the distances between any two friends should be greater than one.
		// This means the friends may be neighbours in at most one of the weeks.
		for(int w1 = 0; w1 < nw-1; w1++){
			for(int w2 = w1+1; w2 < nw; w2++){
				for(int f1 = 0; f1 < nf-1; f1++){
					for(int f2 = f1+1; f2 < nf; f2++){
						model.or(model.arithm(dst[map[w1][f1][f2]], ">", 1),
								 model.arithm(dst[map[w2][f1][f2]], ">", 1)
								).post();
						
					}
				}
			}
		}
		// first-fail search on flattened pos variables
		
		if (first_fail){
			IntVar[] flat = {};
			for (int w = 0; w < nw; w++){
				for (int f = 0; f < nf; f++){
					flat = ArrayUtils.concat(flat, pos[w][f]);
					k = k + 1;
				}
			}
			s.setSearch(minDomLBSearch(flat));
		}
		
		
		if(s.solve()){

			show_table(pos, nw, nf);
			show_distances(dst, map, nw, nf);
		} else {
			System.out.print("no solution");
		}
		stats st = new stats(s);
		st.show();
	}
	
	public static void show_table(IntVar[][] pos, int nw, int nf){
		
		int [][] table = new int [nw][nf];
	
		for(int w = 0; w < nw; w++){ 
			for(int f = 0; f < nf; f++){
				table[w][pos[w][f].getValue()] = f;
			}
		}
		for(int w = 0; w < nw; w++){ 
			System.out.println();
			for(int f = 0; f < nf; f++){
				System.out.print(" "+ table[w][f]);
			}
		}
	}
	
	public static void show_distances(IntVar[] dst, int [][][] map, int nw, int nf){
		System.out.println("\n");
		for(int f1 = 0; f1 < nf-1; f1++){
			for(int f2 = f1+1; f2 < nf; f2++){
				System.out.print(f1 + " & " + f2 + " :");
				for(int w = 0; w < nw; w++){
					System.out.print(" " + dst[map[w][f1][f2]].getValue());
				}
				System.out.println();
			}
		}
	}

}
