package choco;

import org.chocosolver.solver.Model;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.solver.*;

/*
 The problem is to put n balls labelled 1,…,n into 3 boxes so that for any triple of balls
 (x,y,z), with x+y=z, not all are in the same box. This has a solution iff n < 14. 
 
 The problem can be formulated as an 0-1 problem using the variables, Mij for i∈1,…,n,j∈1,2,3
 with Mij true iff ball i is in box j. The constraints are that a ball must be in exactly 
 one box, Mi1+Mi2+Mi3=1 for all i∈1,…,n. And for each x+y=z and j ∈ 1,2,3, not (Mxj∧Myj∧Mzj). 
 This converts to, (1−Mxj)+(1−Myj)+(1−Mzj)≥1 or, Mxj+Myj+Mzj≤2.

 One natural generalization is to consider partitioning into k boxes (for k>3).
*/
public class schurs_lemma {
	
	public static void main(String[] args) {
		int n_boxes = 4;  //     b =  3 / 4
		int n_balls = 40;	// max n = 13 (for b = 3)

		Model m = new Model("(" + n_boxes + "," + n_balls + ") - schur's lemma");
 		Solver s = m.getSolver();
		
 		// Ball i is in Box j iff box[i-1] = j 
		IntVar[] box = m.intVarArray("B", n_balls, 1, n_boxes);

		// Ball i is in the same box of Ball j then the box of Ball i+j is not the same
		for(int i = 1; i <= n_balls; i++){
			for (int j = i; j <= n_balls; j++) {
				if (i+j-1 < n_balls){
					m.or(	m.arithm(box[i-1], "!=", box[j-1]), 
							m.arithm(box[i-1], "!=", box[i+j-1]))
					.post();
				}
			}
		}
				
 		Solution solution = m.getSolver().findSolution();		
		
		if(solution != null){
			System.out.println("\n");
			m.getSolver().printStatistics();
			System.out.println(" --- Statistics ");
			System.out.println("       number of failures = " + String.valueOf(s.getFailCount()));
			System.out.println("     number of backtracks = " + String.valueOf(s.getBackTrackCount()));
			System.out.println("    elapsed cpu time (ms) = " + String.valueOf(s.getTimeCount()*1000));
			System.out.println();
			for (int b = 1; b <= n_boxes; b++){
				System.out.print("box " + b + ":");
				for (int i = 0; i < n_balls; i++){
					if (box[i].getValue() == b){
						System.out.print(" " + (i+1));
					}
				}
				System.out.print("\n");
			}
			
		} else {
			System.out.print("no solutions\n");
		}
		
	}
}